@@ -11,8 +11,9 @@ sub define_shortcut;
 
 BEGIN {
     ## Just leave it on -- with $DB::trace = 0, there doesn't seem
-    ## to be a perforamnce penalty!
-    $^P = 0x303;
+    ## to be a performance penalty!
+    ## enter/exit | per-line | sub locations | eval filenames | anon subnames
+    $^P = 0x313;                # 01 | 02 | 10 | 100 | 200
     $STOPDIE = 1;
     $STOPWARN = 0;
 }
@@ -39,7 +40,8 @@ sub repl_backtrace
     for (my $i = 0; ; ++$i) {
         my ($pack, $file, $line, $sub) = caller($i);
         last unless $pack;
-        print($i == $level+3 ? "*" : ' ', " [$i]\t$sub ($file:$line)\n");
+        # XXX: 4 is the magic number...
+        print($i == $level+4 ? "*" : ' ', " [$i]\t$sub ($file:$line)\n");
     }
 }
 
@@ -57,11 +59,36 @@ sub repl_return
 use vars qw($DIE_TO @DIE_RETURN $DIE_LEVEL);
 $DIE_LEVEL = 0;
 
+sub xreturn
+{
+    eval q{ use Scope::Upper ':all' };
+    if ($@) {
+        print "xreturn requires Sub::Uplevel.\n";
+        return;
+    } else {
+        *xreturn = eval <<'EOS';
+        sub {
+            my $exp = shift;
+            $exp = '""' unless defined $exp;
+            my $ctx = CALLER($level+4); # XXX: ok?
+            local $Sepia::WANTARRAY = want_at $ctx;
+            my @res = eval_in_env($exp, peek_my($level + 4));
+            print STDERR "unwind(@res)\n";
+            unwind @res, SUB UP $ctx;
+        };
+EOS
+        goto &xreturn;
+    }
+}
+
 sub repl_xreturn
 {
-    ($DB::DIE_TO, $DB::DIE_RETURN[0]) = split ' ', $_[0], 2;
-    $DB::DIE_RETURN[0] = $Sepia::REPL{eval}->($DB::DIE_RETURN[0]);
-    last SEPIA_DB_SUB;
+    print STDERR "XRETURN(@_)\n";
+    xreturn(shift);             # XXX: doesn't return.  Problem?
+    print STDERR "XRETURN: XXX\n";
+    # ($DB::DIE_TO, $DB::DIE_RETURN[0]) = split ' ', $_[0], 2;
+    # $DB::DIE_RETURN[0] = $Sepia::REPL{eval}->($DB::DIE_RETURN[0]);
+    # last SEPIA_DB_SUB;
 }
 
 # { package DB;
@@ -117,7 +144,9 @@ sub eval_in_env
         next unless /^([\$\@%])(.+)/;
         $str .= "local *$2 = \$Sepia::ENV->{'$_'}; ";
     }
-    eval "do { no strict; $str $expr }";
+    $str = "do { no strict; package $Sepia::PACKAGE; $str $expr }";
+    return $Sepia::WANTARRAY ? eval $str : scalar eval $str;
+
 }
 
 sub tie_class
@@ -147,9 +176,15 @@ sub eval_in_env2
 }
 
 # evaluate EXP LEV levels up the stack
+#
+# NOTE: We need to act like &repl_eval here and consider e.g. $WANTARRAY
 sub repl_upeval
 {
-    eval_in_env(shift, peek_my(4+$level));
+    # if ($Sepia::WANTARRAY) {
+    return eval_in_env(shift, peek_my(4+$level));
+    # } else {
+    #     return scalar eval_in_env(shift, peek_my(4+$level));
+    # }
 }
 
 # inspect lexicals at level N, or current level
@@ -175,16 +210,13 @@ sub repl_inspect
 sub debug
 {
     my $new = Sepia::as_boolean(shift, $DB::trace);
-    return if $new == $DB::trace;
-    if ($new) {
-        # $^P = 0x2 | 0x10 | 0x100 | 0x200;
-        # *DB::DB = \&repl;
-        $DB::trace = 1;
-        print "debug ON\n";
+    print "debug ", $new ? "ON" : "OFF";
+    if ($new == $DB::trace) {
+        print " (unchanged)\n"
     } else {
-        $DB::trace = 0;
-        print "debug OFF\n";
+        print "\n";
     }
+    $DB::trace = $new;
 }
 
 sub breakpoint_file
@@ -250,6 +282,28 @@ sub repl_delete
     delete $h->{$l} if defined $h;
 }
 
+sub repl_finish
+{
+    # XXX: doesn't handle recursion, but oh, well...
+    my $sub = (caller $level + 4)[3];
+    if (exists $DB::sub{$sub}) {
+        my ($file, $start, $end) = $DB::sub{$sub} =~ /(.*):(\d+)-(\d+)/;
+        print STDERR "finish($sub): will stop at $file:$end\n";
+        # XXX: $end doesn't always work, since it may not have an
+        # executable statement on it.
+        breakpoint($file, $end-1, 'finish');
+        last repl;
+    } else {
+        print STDERR "yikes: @{[keys %DB::sub]}\n";
+    }
+}
+
+sub repl_toplevel
+{
+    local $STOPDIE;
+    die(bless [], __PACKAGE__);
+}
+
 sub add_repl_commands
 {
     define_shortcut 'delete', \&repl_delete,
@@ -258,16 +312,19 @@ sub add_repl_commands
         'debug [0|1]', 'Enable or disable debugging.';
     define_shortcut 'break', \&repl_break,
         'break [F:N [E]]',
-        'Set a breakpoint in F at line N (or at current position), enabled if E evalutes to true.';
+        'Break at file F, line N (or at current position) if E is true.';
     define_shortcut 'lsbreak', \&repl_lsbreak,
         'List breakpoints.';
-    define_shortcut 'dbsub', \&repl_dbsub, '(Un)install DB::sub.';
+    # define_shortcut 'dbsub', \&repl_dbsub, '(Un)install DB::sub.';
     %Sepia::RK = abbrev keys %Sepia::REPL;
 }
 
 sub add_debug_repl_commands
 {
-
+    define_shortcut quit => \&repl_toplevel,
+        'quit', 'Quit the debugger, returning to the top level.';
+    define_shortcut toplevel => \&repl_toplevel,
+        'toplevel', 'Return to the top level.';
     define_shortcut up => sub {
         $level += shift || 1;
         update_location(4);
@@ -297,14 +354,17 @@ sub add_debug_repl_commands
         last repl;
     }, 'step [N]', 'Step N lines forward, entering subroutines.';
 
+    define_shortcut finish => \&repl_finish,
+        'finish', 'Finish the current subroutine.';
+
     define_shortcut list => \&repl_list,
         'list EXPR', 'List source lines of current file.';
     define_shortcut backtrace => \&repl_backtrace, 'show backtrace';
     define_shortcut inspect => \&repl_inspect,
         'inspect [N]', 'inspect lexicals in frame N (or current)';
     define_shortcut return => \&repl_return, 'return EXPR', 'return EXPR';
-    define_shortcut xreturn => \&repl_xreturn, 'xreturn NAME EXPR',
-        'xreturn NAME EXPR';
+    # define_shortcut xreturn => \&repl_xreturn, 'xreturn EXPR',
+    #     'return EXPR from the current sub.';
     define_shortcut eval => \&repl_upeval,
         'eval EXPR', 'evaluate EXPR in current frame';      # DANGER!
 }
@@ -336,6 +396,11 @@ sub DB::DB
         my $cond = $main::{"_<$file"}{$line};
         if ($cond eq 'next') {
             delete $main::{"_<$file"}{$line};
+        } elsif ($cond eq 'finish') {
+            # remove temporary breakpoint and take one more step.
+            delete $main::{"_<$file"}{$line};
+            $DB::single = 1;
+            return;
         } else {
             return unless $Sepia::REPL{eval}->($cond);
         }
